/* This file is part of the 'neper' program. */
/* Copyright (C) 2003-2011, Romain Quey. */
/* See the COPYING file in the top-level directory. */

#include"DeleteEdge.h"

void
UpdateFaceState (struct GEO *pGeo, int delver, int newver)
{
  int i, face;
  int *NewVerFace = ut_alloc_1d_int (100);
  int *DelVerFace = ut_alloc_1d_int (100);

  /* Calculating DelVerFace and NewVerFace */
  neut_geo_ver_face (*pGeo, delver, DelVerFace);
  neut_geo_ver_face (*pGeo, newver, NewVerFace);

  /* every face of the deleted and new vertices will be modified, so
   * their states are set to 1. Maybe some faces will be deleted: in this
   * case, their states will be set to -1 afterwards.
   */
  for (i = 1; i <= DelVerFace[0]; i++)
  {
    face = DelVerFace[i];
    (*pGeo).FaceState[face] = 1;
  }

  for (i = 1; i <= NewVerFace[0]; i++)
  {
    face = NewVerFace[i];
    (*pGeo).FaceState[face] = 1;
  }

  ut_free_1d_int (NewVerFace);
  ut_free_1d_int (DelVerFace);

  return;
}

/* Delete ver from face; note that this routine DOES NOT modify
 * FaceVerQty (it is also the face edge quantity).
 * FaceVerQty is to be modified either in DeleteVerFromFacep
 * or in DeleteEdgeFromFacep, if UpdateFaceVer is after UpdateFaceEdgep
 * and if vice versa, respectively.
 */
void
UpdateFaceVer (struct GEO *pGeo, int delver, int newver, int verbosity)
{
  int i, j;
  int face;
  int *NewVerFace = ut_alloc_1d_int (100);
  int *DelVerFace = ut_alloc_1d_int (100);

  /* Calculating DelVerFace and NewVerFace */
  neut_geo_ver_face (*pGeo, delver, DelVerFace);
  neut_geo_ver_face (*pGeo, newver, NewVerFace);

  for (i = 1; i <= DelVerFace[0]; i++)
  {
    face = DelVerFace[i];

    if (verbosity >= 3)
    {
      ut_print_lineheader (-1);
      printf ("Treating delver face %d (%d/%d) ---\n", face, i,
	      DelVerFace[0]);
      ut_print_lineheader (-1);
      printf ("  %d faces: ", (*pGeo).FaceVerQty[face]);
      for (j = 1; j <= (*pGeo).FaceVerQty[face]; j++)
	printf ("%d ", (*pGeo).FaceVerNb[face][j]);
      printf ("\n");
    }

    if (oneDIntEltPos (NewVerFace, 1, NewVerFace[0], face, 0) != -1)
      DeleteVerFromFace (pGeo, face, delver, verbosity);
    else
      ReplaceVerInFace (pGeo, face, delver, newver);
  }

  ut_free_1d_int (NewVerFace);
  ut_free_1d_int (DelVerFace);

  return;
}

/* Delete edge from face; note that this routine MODIFIES
 * FaceVerQty (it is also the face edge quantity).
 * FaceVerQty is to be modified either in DeleteVerFromFacep
 * or in DeleteEdgeFromFacep, if UpdateFaceVer is after UpdateFaceEdgep
 * and vice versa, respectively.
 */
int
UpdateFaceEdge (struct GEO *pGeo, int edge, int verbosity)
{
  int i;
  int face;
  int err = 0;

  for (i = 0; i <= (*pGeo).EdgeFaceQty[edge] - 1; i++)
  {
    face = (*pGeo).EdgeFaceNb[edge][i];
    if (verbosity >= 3)
    {
      ut_print_lineheader (-1);
      printf ("Treating face %d (%d/%d) ---\n", face, i + 1,
	      (*pGeo).EdgeFaceQty[edge]);
    }

    if ((*pGeo).FaceState[face] == -1)
    {
      ut_print_lineheader (2);
      printf ("The face to remove (%d) is already useless!\n", face);
      abort ();
    }

    if (verbosity >= 3)
    {
      ut_print_lineheader (0);
      printf ("  - face %d contains %d edges: ", face,
	      (*pGeo).FaceVerQty[face]);
    }

    /* if it contains almost 4 edges, the edge is removed;
     * else, the whole face is removed.                     */
    if ((*pGeo).FaceVerQty[face] >= 4)
      DeleteEdgeFromFace (pGeo, face, edge, verbosity);
    else
      err = DeleteFace (pGeo, face, edge, verbosity);
  }

  return err;
}

void
UpdateVerState (struct GEO *pGeo, int delver, int newver, int verbosity)
{
  int wgt1, wgt2;

  if (verbosity >= 3)
  {
    ut_print_lineheader (0);
    printf ("  Updating state of ver %d\n", newver);
    ut_print_lineheader (0);
    printf ("    ver %d: %d and ver %d: %d become ", newver,
	    (*pGeo).VerState[newver], delver, (*pGeo).VerState[delver]);
  }
  /* if state of newver is 0, its wgt is 1, else it is the state */
  wgt1 = IntMax (1, (*pGeo).VerState[newver]);
  /* if state of delver is 0, its wgt is 1, else it is the state */
  wgt2 = IntMax (1, (*pGeo).VerState[delver]);

  /* wgt of newver becomes wgt1+wgt2 */
  (*pGeo).VerState[newver] = wgt1 + wgt2;
  /* state of delver is set to -1 --> useless */
  (*pGeo).VerState[delver] = -1;

  if (verbosity >= 3)
    printf ("ver %d: %d\n", newver, (*pGeo).VerState[newver]);

  return;
}

/* Updating of VerEdgeQty & VerEdgeNb */
void
UpdateVerEdge (struct GEO *pGeo, int edge, int delver, int newver,
	       int verbosity)
{
  int i;
  int *array;			/* will be the VerEdgeNb  of newver ver */
  int qty;			/* will be the VerEdgeQty of newver ver */

  array =
    oneDIntCat ((*pGeo).VerEdgeNb[newver], 0, (*pGeo).VerEdgeQty[newver] - 1,
		(*pGeo).VerEdgeNb[delver], 0, (*pGeo).VerEdgeQty[delver] - 1,
		0);

  if (verbosity >= 3)
  {
    ut_print_lineheader (-1);
    printf ("  Giving ver %d the edges of ver %d\n", newver, delver);

    ut_print_lineheader (-1);
    printf ("    ver %d: ", newver);
    for (i = 0; i <= (*pGeo).VerEdgeQty[newver] - 1; i++)
      printf ("%d ", (*pGeo).VerEdgeNb[newver][i]);

    printf ("and ver %d: ", delver);
    for (i = 0; i <= (*pGeo).VerEdgeQty[delver] - 1; i++)
      printf ("%d ", (*pGeo).VerEdgeNb[delver][i]);
    printf ("become\n");
  }

  qty = (*pGeo).VerEdgeQty[delver] + (*pGeo).VerEdgeQty[newver];
  qty -= oneDIntDeleteNCompress (array, 0, qty - 1, edge, 2);

  (*pGeo).VerEdgeQty[newver] = qty;
  (*pGeo).VerEdgeNb[newver] = ut_realloc_1d_int ((*pGeo).VerEdgeNb[newver], qty);
  ut_array_1d_int_memcpy ((*pGeo).VerEdgeNb[newver], qty, array);

  if (verbosity >= 3)
  {
    ut_print_lineheader (-1);
    printf ("    ver %d: ", newver);
    for (i = 0; i <= (*pGeo).VerEdgeQty[newver] - 1; i++)
      printf ("%d ", (*pGeo).VerEdgeNb[newver][i]);
    printf ("\n");
  }

  ut_free_1d_int (array);

  return;
}

int
UpdateVerBound (struct GEO *pGeo, int delver, int newver, int verbosity)
{
  int i;
  int qty;
  int *array;
  int err = 0;

  if (verbosity >= 3)
  {
    ut_print_lineheader (-1);
    printf ("  Giving ver %d the boundaries of ver %d\n", newver, delver);

    ut_print_lineheader (-1);
    printf ("ver %d: ", delver);
    for (i = 1; i <= (*pGeo).VerBound[delver][0]; i++)
      printf ("%d ", (*pGeo).VerBound[delver][i]);
    printf ("and ver %d: ", newver);
    for (i = 1; i <= (*pGeo).VerBound[newver][0]; i++)
      printf ("%d ", (*pGeo).VerBound[newver][i]);
    printf ("becomes: ver %d: ", newver);
  }

  qty = (*pGeo).VerBound[delver][0] + (*pGeo).VerBound[newver][0];

  if (qty != 0)
  {
    array =
      oneDIntCat ((*pGeo).VerBound[delver], 1, (*pGeo).VerBound[delver][0],
		  (*pGeo).VerBound[newver], 1, (*pGeo).VerBound[newver][0],
		  0);

    /* oneDIntSort(array,0,qty-1); */
    gsl_sort_int (array, 1, qty);
    qty = oneDIntCompress (array, 0, qty - 1) + 1;

    if ((oneDIntEltPos (array, 0, qty - 1, -1, 0) != -1 &&
	 oneDIntEltPos (array, 0, qty - 1, -2, 0) != -1)
	|| (oneDIntEltPos (array, 0, qty - 1, -3, 0) != -1 &&
	    oneDIntEltPos (array, 0, qty - 1, -4, 0) != -1)
	|| (oneDIntEltPos (array, 0, qty - 1, -5, 0) != -1 &&
	    oneDIntEltPos (array, 0, qty - 1, -6, 0) != -1))
    {
      /* ut_print_messagewnc(2,72,"cl is too large!"); */
      /* abort();                             */
      err = 1;
    }

    (*pGeo).VerBound[newver][0] = qty;
    for (i = 1; i <= (*pGeo).VerBound[newver][0]; i++)
      (*pGeo).VerBound[newver][i] = array[i - 1];

    ut_free_1d_int (array);
  }

  if (verbosity >= 3)
  {
    for (i = 1; i <= (*pGeo).VerBound[newver][0]; i++)
      printf ("%d ", (*pGeo).VerBound[newver][i]);
    printf ("\n");
  }

  return err;
}

void
UpdateEdgeLength (struct GEO *pGeo, int ver)
{
  int i;

  for (i = 0; i < (*pGeo).VerEdgeQty[ver]; i++)
    neut_geo_edge_length (pGeo, (*pGeo).VerEdgeNb[ver][i]);

  return;
}

void
UpdateVerCooBary (struct GEO *pGeo, int delver, int newver)
{
  int i;
  double wgt1, wgt2;

  /* if state of newver is 0, its wgt is 1, else it is the state */
  wgt1 = (double) IntMax (1, (*pGeo).VerState[newver]);
  /* if state of delver is 0, its wgt is 1, else it is the state */
  wgt2 = (double) IntMax (1, (*pGeo).VerState[delver]);

  /* We consider the barycenter */
  for (i = 0; i <= 2; i++)
    (*pGeo).VerCoo[newver][i] = (wgt1 * (*pGeo).VerCoo[newver][i] +
				 wgt2 * (*pGeo).VerCoo[delver][i]) / (wgt1 +
								      wgt2);

  /* but if one of the parent vertices belongs to one/more boundaries, */
  /* the newver vertex is put on this/these boundaries. */
  for (i = 1; i <= (*pGeo).VerBound[newver][0]; i++)
    switch ((*pGeo).VerBound[newver][i])
    {
    case -1:
      (*pGeo).VerCoo[newver][0] = (*pGeo).DomainSize[0][0];	/* x=0 */
      break;
    case -2:
      (*pGeo).VerCoo[newver][0] = (*pGeo).DomainSize[0][1];	/* x=max */
      break;
    case -3:
      (*pGeo).VerCoo[newver][1] = (*pGeo).DomainSize[1][0];	/* y=0 */
      break;
    case -4:
      (*pGeo).VerCoo[newver][1] = (*pGeo).DomainSize[1][1];	/* y=max */
      break;
    case -5:
      (*pGeo).VerCoo[newver][2] = (*pGeo).DomainSize[2][0];	/* z=0 */
      break;
    case -6:
      (*pGeo).VerCoo[newver][2] = (*pGeo).DomainSize[2][1];	/* z=max */
      break;
    default:
      break;
    }

  return;
}

void
UpdateVerCooMiniFF (struct GEO *pGeo, int newver, int verbosity)
{
  int f, i, j;
  double **A;
  double *B;
  double *X;
  int *face = ut_alloc_1d_int (100);
  int tmpface;

  A = ut_alloc_2d (4, 4);
  B = ut_alloc_1d (4);
  X = ut_alloc_1d (4);

  neut_geo_ver_face (*pGeo, newver, face);

  for (f = 1; f <= face[0]; f++)
  {
    tmpface = face[f];

    for (i = 1; i <= 3; i++)
    {
      for (j = 1; j <= 3; j++)
	A[i][j] += (*pGeo).FaceEq[tmpface][i] * (*pGeo).FaceEq[tmpface][j];

      B[i] += (*pGeo).FaceEq[tmpface][i] * (*pGeo).FaceEq[tmpface][0];
    }
  }

  /* but if one of the parent vertices belongs to one/more boundaries, */
  /* the newver vertex is put on this/these boundaries. */
  for (i = 1; i <= (*pGeo).VerBound[newver][0]; i++)
  {
    if (verbosity >= 3)
      printf ("bound: %d/%d: %d\n", i, (*pGeo).VerBound[newver][0],
	      (*pGeo).VerBound[newver][i]);

    switch ((*pGeo).VerBound[newver][i])
    {
    case -1:			/* x=0 */
      A[1][1] = 1;
      A[1][2] = 0;
      A[1][3] = 0;
      B[1] = (*pGeo).DomainSize[0][0];
      break;
    case -2:			/* x=max */
      A[1][1] = 1;
      A[1][2] = 0;
      A[1][3] = 0;
      B[1] = (*pGeo).DomainSize[0][1];
      break;
    case -3:			/* y=0 */
      A[2][1] = 0;
      A[2][2] = 1;
      A[2][3] = 0;
      B[2] = (*pGeo).DomainSize[1][0];
      break;
    case -4:			/* y=max */
      A[2][1] = 0;
      A[2][2] = 1;
      A[2][3] = 0;
      B[2] = (*pGeo).DomainSize[1][1];
      break;
    case -5:			/* z=0 */
      A[3][1] = 0;
      A[3][2] = 0;
      A[3][3] = 1;
      B[3] = (*pGeo).DomainSize[2][0];
      break;
    case -6:			/* z=max */
      A[3][1] = 0;
      A[3][2] = 0;
      A[3][3] = 1;
      B[3] = (*pGeo).DomainSize[2][1];
      break;
    default:
      /*printf("VerBound: qty:%d: %dth=%d\n",(*pGeo).VerBound[newver][0],i,(*pGeo).VerBound[newver][i]); */
      /*abort(); */
      break;
    }
  }


  SolveMatrix (3, A, X, B);

  for (i = 0; i <= 2; i++)
    (*pGeo).VerCoo[newver][i] = X[i + 1];

  if (verbosity >= 3)
  {
    if ((*pGeo).VerBound[newver][0] != 0)
    {
      printf ("bound: %d: ", (*pGeo).VerBound[newver][0]);
      for (i = 1; i <= (*pGeo).VerBound[newver][0]; i++)
	printf ("%d ", (*pGeo).VerBound[newver][i]);
      printf ("\ncoo: ");
      for (i = 0; i <= 2; i++)
	printf ("%f ", (*pGeo).VerCoo[newver][i]);
      printf ("\nA:\n");
      for (i = 1; i <= 3; i++)
      {
	for (j = 1; j <= 3; j++)
	  printf ("%f ", A[i][j]);
	printf ("\n");
      }
    }
  }

  ut_free_2d (A, 4);
  ut_free_1d (B);
  ut_free_1d (X);
  ut_free_1d_int (face);

  return;
}
