/**********************************************************************
 * Orilib - routines for orientation manipulation.                    *
 * Copyright (C) 2007, 2008, 2009 Romain Quey                         *
 *                                                                    *
 * This program is free software: you can redistribute it and/or      *
 * modify it under the terms of the GNU General Public License as     *
 * published by the Free Software Foundation, either version 3 of the *
 * License, or (at your option) any later version.                    *
 *                                                                    *
 * This program is distributed in the hope that it will be useful,    *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of     *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  *
 * General Public License for more details.                           *
 *                                                                    *
 * You should have received a copy of the GNU General Public License  *
 * along with this program.  If not, see                              *
 * <http://www.gnu.org/licenses/>.                                    *
 *                                                                    *
 * project homepage: http://orilib.sourceforge.net                    *
 **********************************************************************/
#include"ol_set.h"

void
ol_set_mean (struct OL_SET set, char* crysym, double *qm)
{
  if (set.size > 0)
    ol_set_mean_ref (set, crysym, set.q[0], qm);

  return;
}

void
ol_set_mean_ref (struct OL_SET set, char* crysym, double *qref, double *qm)
{
  unsigned int i, j;
  double qnorm;
  double *qb = ol_q_alloc ();

  ut_array_1d_zero (qm, 4);

  for (i = 0; i < set.size; i++)
  {
    ol_q_q_qdisori_cur (qref, set.q[i], crysym, qb);

    for (j = 0; j < 4; j++)
      qm[j] += qb[j] * set.weight[i];
  }

  qnorm = ut_array_1d_norm (qm, 4);
  ut_array_1d_scale (qm, 4, 1 / qnorm);

  ol_q_q_q (qref, qm, qm);

  ol_q_free (qb);

  return;
}

int
ol_set_mean_iter (struct OL_SET set, char* crysym, double *qm)
{
  double *qb = ol_q_alloc ();
  double theta;
  int it, res = 0;

  it = 0;
  ol_q_memcpy (set.q[0], qb);

  do
  {
    ol_set_mean_ref (set, crysym, qb, qm);
    ol_q_q_disori (qb, qm, crysym, &theta);
    ol_q_memcpy (qm, qb);
    it++;
  }
  while (theta > 1e-1 && it <= 100);

  if (it > 100)
    res = -1;

  ol_q_free (qb);

  return res;
}

int
ol_set_mean_iter_ref (struct OL_SET set, char* crysym, double *qref, double *qm)
{
  double *qb = ol_q_alloc ();
  double theta;
  int it, res = 0;

  it = 0;
  ol_q_memcpy (qref, qb);

  do
  {
    ol_set_mean_ref (set, crysym, qb, qm);
    ol_q_q_disori (qb, qm, crysym, &theta);
    ol_q_memcpy (qm, qb);
    it++;
  }
  while (theta > 1e-1 && it <= 100);

  if (it > 100)
    res = -1;

  ol_q_free (qb);

  return res;
}

struct OL_SET
ol_set_alloc (size_t size)
{
  unsigned int i;
  struct OL_SET Set;

  Set.size = size;
  Set.q = ut_alloc_2d (size, 4);
  Set.weight = ut_alloc_1d (size);
  for (i = 0; i < size; i++)
    Set.weight[i] = 1;

  return Set;
}

void
ol_set_free (struct OL_SET Set)
{
  ut_free_2d (Set.q, Set.size);
  ut_free_1d (Set.weight);

  return;
}

int
ol_set_fscanf (FILE * file, struct OL_SET *pSet, int *pf)
{
  size_t i;
  double *e = ol_e_alloc ();
  unsigned int size;
  int test, firstline_nbw, whole_nbw;

  // checking the number of string on the first line
  // can be 1, 3 or 4. 
  firstline_nbw = ut_file_line_nbwords_pointer (file);
  whole_nbw = ut_file_nbwords_pointer (file);

  /* required because size_t cannot be scanf (???) neither
     with %u nor with %zu with -pedantic (ISO C90) */
  /* adding 'z' is suggested on
     http://manpages.ubuntu.com/manpages/jaunty/man3/scanf.3.html */

  if (firstline_nbw == 1)
  {
    test = fscanf (file, "%u", &size);
    (*pSet).size = (size_t) size;
    if (pf != NULL)
      (*pf) = 3;

    if (test == -1)
      return 0;
  }
  else if (firstline_nbw == 3 || firstline_nbw == 4)
    (*pSet).size = whole_nbw / firstline_nbw;
  else
    return 0;

  if (pf != NULL)
  {
    if (firstline_nbw == 3)
      (*pf) = 1;
    else if (firstline_nbw == 4)
      (*pf) = 2;
  }

  (*pSet) = ol_set_alloc ((*pSet).size);

  for (i = 0; i < (*pSet).size; i++)
  {
    ol_e_fscanf (file, e);
    ol_e_q (e, (*pSet).q[i]);

    if (firstline_nbw != 3)
      fscanf (file, "%lf", &((*pSet).weight[i]));
  }

  ol_e_free (e);

  return 1;
}

int
ol_set_fscanf_short (FILE * file, struct OL_SET *pSet)
{
  size_t i;
  int test = 0;
  double *e = ol_e_alloc ();

  (*pSet).size = ut_file_nblines_pointer (file);

  (*pSet) = ol_set_alloc ((*pSet).size);

  for (i = 0; i < (*pSet).size; i++)
  {
    test = ol_e_fscanf (file, e);
    ol_e_q (e, (*pSet).q[i]);
  }

  ol_e_free (e);

  return test;
}

int
ol_set_fscanf_sample (FILE * file, double factor, struct OL_SET *pSet, int* pf)
{
  /* needed for -O3 opti */
  struct OL_SET Setb = ol_set_alloc (1);

  ol_set_fscanf (file, &Setb, pf);
  ol_set_sample (Setb, factor, pSet);
  ol_set_free (Setb);

  return 1;
}

int
ol_set_fscanf_sample_short (FILE * file, double factor, struct OL_SET *pSet)
{
  /* needed for -O3 opti */
  struct OL_SET Setb = ol_set_alloc (1);

  ol_set_fscanf_short (file, &Setb);
  ol_set_sample (Setb, factor, pSet);
  ol_set_free (Setb);

  return 1;
}

int
ol_set_fprintf (FILE * file, struct OL_SET Set, int f, char *format)
{
  unsigned int i;
  int test;
  double *e = ol_e_alloc ();
  char *fmt = ut_alloc_1d_char (4 * strlen (format) + 10);

  test = -1;

  if (f == 3)
  {
    test = fprintf (file, "%u\n", (unsigned int) Set.size);
    sprintf (fmt, "%s %s %s\n", format, format, format);
  }
  else if (f == 2)
    sprintf (fmt, "%s %s %s %s\n", format, format, format, format);
  else if (f == 1)
    sprintf (fmt, "%s %s %s\n", format, format, format);
  else
    abort ();

  if (f == 1)
    for (i = 0; i < Set.size; i++)
    {
      ol_q_e (Set.q[i], e);
      test = fprintf (file, fmt, e[0], e[1], e[2]);
    }
  else
    for (i = 0; i < Set.size; i++)
    {
      ol_q_e (Set.q[i], e);
      test = fprintf (file, fmt, e[0], e[1], e[2], Set.weight[i]);
    }

  ol_e_free (e);
  ut_free_1d_char (fmt);

  return test;
}

int
ol_set_fprintf_short (FILE * file, struct OL_SET Set, char *format)
{
  unsigned int i;
  int test = 0;
  double *e = ol_e_alloc ();

  for (i = 0; i < Set.size; i++)
  {
    ol_q_e (Set.q[i], e);
    test = ol_e_fprintf (file, e, format);
  }

  ol_e_free (e);

  return test;
}

void
ol_set_subset (struct OL_SET Set1, double factor, struct OL_SET *pSet2)
{
  unsigned int qty, id;
  double factorinv;

  if (factor < 1e-6)
    return;

  factorinv = 1 / factor;

  qty = (unsigned int) ((double) Set1.size * factor) + 2;

  (*pSet2).size = 0;
  (*pSet2).q = ut_alloc_2d (qty, 4);
  (*pSet2).weight = ut_alloc_1d (qty);

  id = (int) (factorinv * .5 - .5);
  while (id < Set1.size)
  {
    ol_q_memcpy (Set1.q[id], (*pSet2).q[(*pSet2).size]);
    (*pSet2).weight[(*pSet2).size] = Set1.weight[id];
    (*pSet2).size++;
    id = (int) (factorinv * ((*pSet2).size + 0.5) - 0.5);
  }

  return;
}

void
ol_set_orthotropic (struct OL_SET Set1, struct OL_SET *pSet2)
{
  unsigned int i, j, id;
  double **qsym = ut_alloc_2d (3, 4);

  (*pSet2) = ol_set_alloc (Set1.size * 4);

  id = 0;
  for (i = 0; i < Set1.size; i++)
  {
    ol_q_refsym (Set1.q[i], 1, qsym[0]);
    ol_q_refsym (Set1.q[i], 2, qsym[1]);
    ol_q_refsym (qsym[0], 2, qsym[2]);

    ol_q_memcpy (Set1.q[i], (*pSet2).q[id++]);
    for (j = 0; j < 3; j++)
      ol_q_memcpy (qsym[j], (*pSet2).q[id++]);
  }

  ut_free_2d (qsym, 3);

  return;
}
  
void
ol_set_memcpy (struct OL_SET Set1, struct OL_SET* pSet2)
{
  (*pSet2).size = Set1.size;
  ut_array_2d_memcpy ((*pSet2).q, Set1.size, 4, Set1.q);
  ut_array_1d_memcpy ((*pSet2).weight, Set1.size, Set1.weight);

  return;
}

void
ol_set_aniso (struct OL_SET set, double **evect, double *eval)
{
  int i, j, k;
  int n;
  double **R = ut_alloc_2d (set.size, 3);
  double** evect2 = ol_g_alloc ();

  double **ERR = ut_alloc_2d (3, 3);
  double *ER = ut_alloc_1d (3);
  double **Sb = ut_alloc_2d (3, 3);
  double* theta = ut_alloc_1d (4);
  double sumweight;

  if (set.size == 0)
  {
    ut_print_message (2, 0, "ol_set_aniso: set of size 0");
    abort ();
  }
  else if (set.size == 1)
  {
    ol_g_set_id (evect);
    ut_array_1d_set (eval, 3, 0);
  }
  else
  {
    n = set.size;
    for (i = 0; i < n; i++)
      ol_q_R (set.q[i], R[i]);

    /* calculation of the covariant matrix S */

    if (set.weight != NULL)
    {
      sumweight = ut_array_1d_sum (set.weight, n);
      for (i = 0; i < 3; i++)
	for (j = 0; j < 3; j++)
	{
	  for (k = 0; k < n; k++)
	    ERR[i][j] += set.weight[k] * R[k][i] * R[k][j];

	  ERR[i][j] /= sumweight;
	}

      for (i = 0; i < 3; i++)
      {
	for (k = 0; k < n; k++)
	  ER[i] += set.weight[k] * R[k][i];

	ER[i] /= sumweight;
      }
    }
    else
    {
      for (i = 0; i < 3; i++)
	for (j = 0; j < 3; j++)
	{
	  for (k = 0; k < n; k++)
	    ERR[i][j] += R[k][i] * R[k][j];

	  ERR[i][j] /= (double) n;
	}

      for (i = 0; i < 3; i++)
      {
	for (k = 0; k < n; k++)
	  ER[i] += R[k][i];

	ER[i] /= (double) n;
      }
    }

    for (i = 0; i < 3; i++)
      for (j = 0; j < 3; j++)
	Sb[i][j] = ERR[i][j] - ER[i] * ER[j];

    /* calculation of the eigen values and vectors of S */

    ut_mat_eigen (3, Sb, eval, evect);

    /* 114.59... is "(180 / PI) * 2" */
    for (i = 0; i < 3; i++)
      eval[i] = 114.59155902616464175442 * atan (ut_num_sqrt (eval[i]));

    // v2 is recalculated as v0 'vect prod' v1 to get a right-handed
    // csys.

    ut_vector_vectprod (evect[0], evect[1], evect[2]);

    // taking the 4 mmm symmetries of evect, and keeping the less
    // disoriented one.
    /*
    for (i = 1; i <= 4; i++)
    {
      ol_g_crysym (evect, "mmm", i, evect2);
      ol_g_theta (evect2, &(theta[i - 1]));
      printf ("i = %d, theta = %f\n", i, theta[i - 1]);
    }
    min_id = 1 + ut_array_1d_min_index (theta, 4);

    ol_g_crysym (evect, "mmm", min_id, evect);
    */
  }

  ut_free_2d (R, set.size);
  ol_g_free (evect2);
  ut_free_1d (theta);
  ut_free_2d (ERR, 3);
  ut_free_1d (ER);
  ut_free_2d (Sb,  3);

  return;
}

void
ol_q_aniso_theta (double *q, double **evect, double *theta)
{
  int j, k;
  double tmp;
  double *R = ol_R_alloc ();

  ol_q_R (q, R);

  /* Calculation of the thetas */
  for (j = 0; j < 3; j++)
  {
    // printf ("calc theta along ");
    // ut_array_1d_fprintf (stdout, evect[j], 3, "%9.6f");
    tmp = 0;
    for (k = 0; k < 3; k++)
      tmp += evect[j][k] * R[k];

    /* 114.59155902616464175442 is (180 / PI) * 2 */
    theta[j] = 114.59155902616464175442 * atan (tmp);
  }

  ol_R_free (R);

  return;
}

void
ol_set_aniso_thetadistrib (struct OL_SET set, double **evect, double step, double **pdistrib, double *distribfirstx, int *distriblength)
{
  int i, j;
  int n = set.size;
  double **theta = ut_alloc_2d (3, n);
  double *etheta = ut_alloc_1d (3);
  double* q = ol_q_alloc ();

  /* Calculation of the thetas */
  for (i = 0; i < n; i++)
  {
    ol_q_aniso_theta (set.q[i], evect, etheta);

    for (j = 0; j < 3; j++)
      theta[j][i] = etheta[j];
  }

  /* Allocation and calculation */
  for (j = 0; j < 3; j++)
  {
    ut_array_distribparam (theta[j], n, step,
			   &(distribfirstx[j]), &(distriblength[j]));

    pdistrib[j] = ut_alloc_1d (distriblength[j]);

    ut_array_distrib (theta[j], n, step, distribfirstx[j],
		      distriblength[j], pdistrib[j]);
  }

  ut_free_2d (theta, 3);
  ut_free_1d (etheta);
  ol_q_free (q);

  return;
}

void
ol_set_sample (struct OL_SET Set1, double factor, struct OL_SET *pSet2)
{
  unsigned int qty, id;
  double factorinv = 1 / factor;

  qty = (unsigned int) ((double) Set1.size * factor) + 10;

  (*pSet2).size = 0;
  (*pSet2).q = ut_alloc_2d (qty, 4);
  (*pSet2).weight = ut_alloc_1d (qty);

  id = (unsigned int) (factorinv * .5 - .5);
  while (id < Set1.size)
  {
    ol_q_memcpy (Set1.q[id], (*pSet2).q[(*pSet2).size]);
    (*pSet2).weight[(*pSet2).size] = Set1.weight[id];
    (*pSet2).size++;
    id = (unsigned int) (factorinv * ((*pSet2).size + 0.5) - 0.5);
  }

  return;
}

void
ol_set_disori (struct OL_SET set, char* crysym, double *qmean, double *pdisori)
{
  unsigned int i;
  double tmp;

  (*pdisori) = 0;
  for (i = 0; i < set.size; i++)
  {
    ol_q_q_disori (set.q[i], qmean, crysym, &tmp);
    (*pdisori) += tmp;
  }
  (*pdisori) /= set.size;

  return;
}

void
ol_set_disoriset (struct OL_SET set1, char* crysym, struct OL_SET* psetd)
{
  unsigned int i;
  double *qm = ol_q_alloc ();

  ol_set_mean_iter (set1, crysym, qm);
  
  (*psetd).size = set1.size;
  for (i = 0; i < set1.size; i++)
  {
    ol_q_q_qdisori (qm, set1.q[i], crysym, (*psetd).q[i]);
    (*psetd).weight[i] = set1.weight[i];
  }

  ol_q_free (qm);

  return;
}

void
ol_set_mean_disoriset (struct OL_SET set1, double* qm, char* crysym, struct OL_SET* psetd)
{
  unsigned int i;
  
  (*psetd).size = set1.size;
  for (i = 0; i < set1.size; i++)
  {
    ol_q_q_qdisori (qm, set1.q[i], crysym, (*psetd).q[i]);
    (*psetd).weight[i] = set1.weight[i];
  }

  return;
}

int
ol_set_filter (struct OL_SET set1, char* crysym, double rel, struct OL_SET *pset2, double* qmean, double* pdis)
{
  unsigned int i;
  double *qm = ol_q_alloc ();
  double mdis;
  double *dis = ut_alloc_1d (set1.size);

  if (set1.size == 0)
  {
    ut_print_message (2, 0, "ol_set_filter: input set of size 0\n");
    abort ();
  }
  else if (set1.size == 1)
  {
    ol_set_memcpy (set1, pset2);
    return 0;
  }

  ol_set_mean (set1, crysym, qm);

  mdis = 0;
  for (i = 0; i < set1.size; i++)
  {
    ol_q_q_disori (qm, set1.q[i], crysym, &dis[i]);
    mdis += dis[i];
  }
  mdis *= (rel / set1.size);

  (*pset2).size = 0;
  for (i = 0; i < set1.size; i++)
    if (dis[i] < mdis + OL_EPS_DEG)
      ol_q_memcpy (set1.q[i], (*pset2).q[(*pset2).size++]);

  ol_q_free (qm);
  ut_free_1d (dis);

  if (pdis != NULL)
    (*pdis) = mdis;
  if (qmean != NULL)
    ol_q_memcpy (qm, qmean);

  if ((*pset2).size == 0)
  {
    ol_set_memcpy (set1, pset2);
    return -1;
  }
  else
    return set1.size - (*pset2).size;
}

int
ol_set_filter_iter (struct OL_SET set1, char* crysym, double rel, struct OL_SET *pset2, double* qmean, double* pdis)
{
  ol_set_filter (set1, crysym, rel, pset2, qmean, pdis);

  while (ol_set_filter (*pset2, crysym, rel, pset2, qmean, pdis) > 0)
    continue;

  if ((*pset2).size == 0)
    return -1;
  else
    return set1.size - (*pset2).size;
}

int
ol_set_anifilter (struct OL_SET set1, char* crysym, double rel, struct OL_SET *pset2)
{
  unsigned int i;
  double *qm = ol_q_alloc ();
  double **v = ut_alloc_2d (3, 3);
  double *mdis = ut_alloc_1d (3);
  double *dis = ut_alloc_1d (3);
  struct OL_SET setd = ol_set_alloc (set1.size);

  if (set1.size == 0)
  {
    ut_print_message (2, 0, "ol_set_filter: input set of size 0\n");
    abort ();
  }

  ol_set_disoriset (set1, crysym, &setd);
  ol_set_aniso (setd, v, mdis);
  ut_array_1d_scale (mdis, 3, rel);

  (*pset2).size = 0;
  for (i = 0; i < set1.size; i++)
  {
    ol_q_aniso_theta (setd.q[i], v, dis);
    if (fabs (dis[0]) < mdis[0] + OL_EPS_DEG 
     && fabs (dis[1]) < mdis[1] + OL_EPS_DEG
     && fabs (dis[2]) < mdis[2] + OL_EPS_DEG)
    {
      ol_q_memcpy (set1.q[i], (*pset2).q[(*pset2).size]);
      (*pset2).size++;
    }
  }

  ol_q_free (qm);
  ut_free_2d (v, 3);
  ut_free_1d (mdis);
  ut_free_1d (dis);
  ol_set_free (setd);

  if ((*pset2).size == 0)
  {
    ol_set_memcpy (set1, pset2);
    return -1;
  }
  else
    return set1.size - (*pset2).size;
}

int
ol_set_mean_anifilter (struct OL_SET set1, double* qm, char* crysym, double rel, struct OL_SET *pset2)
{
  unsigned int i, j;
  double **v = ut_alloc_2d (3, 3);
  double *mdis = ut_alloc_1d (3);
  double *relmdis = ut_alloc_1d (3);
  double *dis = ut_alloc_1d (3);
  struct OL_SET setd = ol_set_alloc (set1.size);
  double radius;

  if (set1.size == 0)
  {
    ut_print_message (2, 0, "ol_set_filter: input set of size 0\n");
    abort ();
  }

  ol_set_mean_disoriset (set1, qm, crysym, &setd);
  ol_set_aniso (setd, v, mdis);

  for (i = 0; i < 3; i++)
    relmdis[i] = rel * mdis[i];

  (*pset2).size = 0;
  for (i = 0; i < set1.size; i++)
  {
    ol_q_aniso_theta (setd.q[i], v, dis);

    radius = 0;
    for (j = 0; j < 3; j++)
      radius += pow (dis[j] / relmdis[j], 2);
    if (radius < 1)
    {
      ol_q_memcpy (set1.q[i], (*pset2).q[(*pset2).size]);
      (*pset2).size++;
    }
  }

  ut_free_2d (v, 3);
  ut_free_1d (mdis);
  ut_free_1d (relmdis);
  ut_free_1d (dis);
  ol_set_free (setd);

  if ((*pset2).size == 0)
  {
    ol_set_memcpy (set1, pset2);
    return -1;
  }
  else
    return set1.size - (*pset2).size;
}


int
ol_set_mean_aniso_anifilter (struct OL_SET set1, double* qm, double** v, double* mdis, char* crysym, double rel, struct OL_SET *pset2, struct OL_SET *pset3)
{
  unsigned int i, j;
  double radius;
  double *relmdis = ut_alloc_1d (3);
  double *dis = ut_alloc_1d (3);
  struct OL_SET setd;

  if (set1.size == 0)
  {
    ut_print_message (2, 0, "Input set of size 0\n");
    abort ();
  }

  setd = ol_set_alloc (set1.size);
  ol_set_mean_disoriset (set1, qm, crysym, &setd);

  for (i = 0; i < 3; i++)
    relmdis[i] = rel * mdis[i];

  (*pset2).size = 0;

  if (pset3 != NULL)
    (*pset3).size = 0;

  for (i = 0; i < set1.size; i++)
  {
    ol_q_aniso_theta (setd.q[i], v, dis);
    radius = 0;
    for (j = 0; j < 3; j++)
      radius += (dis[j] / relmdis[j]) * (dis[j] / relmdis[j]);
    if (radius < 1)
    {
      ol_q_memcpy (set1.q[i], (*pset2).q[(*pset2).size]);
      (*pset2).weight[(*pset2).size] = set1.weight[i], 
      (*pset2).size++;
    }
    else if (pset3 != NULL)
    {
      ol_q_memcpy (set1.q[i], (*pset3).q[(*pset3).size]);
      (*pset3).weight[(*pset3).size] = set1.weight[i], 
      (*pset3).size++;
    }
  }

  ut_free_1d (relmdis);
  ut_free_1d (dis);
  ol_set_free (setd);

  if ((*pset2).size == 0)
  {
    ol_set_memcpy (set1, pset2);
    return -1;
  }
  else
    return set1.size - (*pset2).size;
}

int
ol_set_mean_aniso_anifilter_iter (struct OL_SET set1, double* qm0,
    double** v0, double* mdis0, char* crysym, double rel, struct OL_SET
    *pset2, double* qm, double** v, double* mdis, struct OL_SET* pset3)
{
  return ol_set_mean_aniso_anifilter_iter_itermax (set1, qm0, v0, mdis0,
      crysym, rel, OL_ITERMAX2, pset2, qm, v, mdis, pset3);
}


int
ol_set_anifilter_iter (struct OL_SET set1, char* crysym, double rel,
			  struct OL_SET *pset2, double* qm,
			  double** v, double* mdis, struct OL_SET* pset3)
{
  return ol_set_anifilter_iter_itermax (set1, crysym, rel, OL_ITERMAX2,
      pset2, qm, v, mdis, pset3);
}


int
ol_set_anifilter_iter_itermax (struct OL_SET set1, char* crysym, double rel,
			  int maxiter, struct OL_SET *pset2, double* qm,
			  double** v, double* mdis, struct OL_SET *pset3)
{
  return ol_set_mean_aniso_anifilter_iter_itermax (set1, NULL, NULL,
      NULL, crysym, rel, maxiter, pset2, qm, v, mdis, pset3);
}


int
ol_set_mean_aniso_anifilter_iter_itermax (struct OL_SET set1, double*
    qmean0, double** v0, double* mdis0, char* crysym, double rel, int
    maxiter, struct OL_SET *pset2, double* qm, double** v, double* mdis,
    struct OL_SET *pset3)
{
  int i;
  double* q = ol_q_alloc ();
  double* qmean = ol_q_alloc ();
  double* r = ol_r_alloc ();
  double theta;
  double** gshake = ol_g_alloc ();
  double* mdis_cur = ut_alloc_1d (3);
  double** v_cur = ol_g_alloc ();
  double* mdis_prev = ut_alloc_1d (3);
  double** v_prev = ol_g_alloc ();
  struct OL_SET Setd;
  struct OL_SET set1cpy = ol_set_alloc (set1.size);
  ol_set_memcpy (set1, &set1cpy);

  maxiter = 2;

  ol_set_memcpy (set1cpy, pset2);
  Setd = ol_set_alloc (set1cpy.size);

  ut_array_1d_set (mdis_prev, 3, 123);
  ol_g_set_id (v_prev);

  for (i = 1; i <= maxiter; i++)
  {
    // calculating mean
    if (i == 1)
    {
      if (qmean0 != NULL)
	ol_q_memcpy (qmean0, qmean);
      else
	ol_set_mean_iter ((*pset2), crysym, qmean);
    }
    else
      ol_set_mean_iter_ref ((*pset2), crysym, qmean, qmean);

    ol_set_mean_disoriset ((*pset2), qmean, crysym, &Setd);

    if (i == 1 && v0 != NULL && mdis0 != NULL)
    {
      ol_g_memcpy (v0, v_cur);
      ol_e_memcpy (mdis, mdis_cur);
    }
    else
      ol_set_aniso (Setd, v_cur, mdis_cur);

    if (i % 50 != 0) // standard case
      ol_set_mean_aniso_anifilter (set1cpy, qmean, v_cur, mdis_cur, crysym, rel, pset2, pset3);
    else // fastening convergence
    {
      ol_nb_max_rtheta (drand48(), drand48(), drand48(), 2 * 50 / i * mdis_cur[2], r, &theta);
      ol_rtheta_q (r, theta, q);
      ol_q_q_q (qmean, q, qmean);

      ol_set_mean_aniso_anifilter (set1cpy, qmean, v_cur, mdis_cur, crysym, rel, pset2, pset3);

      while ((*pset2).size == 0)
      {
	ol_nb_max_rtheta (drand48(), drand48(), drand48(), 2 * 50 / i * mdis_cur[2], r, &theta);
	ol_rtheta_g (r, theta, gshake);
	ol_g_g_g (v_cur, gshake, v_cur);
	ol_set_mean_aniso_anifilter (set1cpy, qmean, v_cur, mdis_cur, crysym, rel, pset2, pset3);
      }
    }

    // testing convergence
    ol_g_g_disori (v_cur, v_prev, "mmm", &theta);
    if (theta < 2)
      if (2 * fabs (mdis_cur[0] - mdis_prev[0])/(mdis_cur[0] + mdis_prev[0]) < 0.01 
       && 2 * fabs (mdis_cur[1] - mdis_prev[1])/(mdis_cur[1] + mdis_prev[1]) < 0.01 
       && 2 * fabs (mdis_cur[2] - mdis_prev[2])/(mdis_cur[2] + mdis_prev[2]) < 0.01)
      break;

    ol_r_memcpy (mdis_cur, mdis_prev);
    ol_g_memcpy (v_cur, v_prev);
  }

  if (qm != NULL)
    ol_q_memcpy (qmean, qm);
  if (v != NULL)
    ol_g_memcpy (v_cur, v);
  if (mdis != NULL)
    ol_r_memcpy (mdis_cur, mdis);

  ol_set_aniso (Setd, v, mdis);

  ol_set_free (Setd);
  ol_set_free (set1cpy);

  return (i < maxiter) ? i : -1;
}

void
ol_set_aniso_modes (struct OL_SET Set, double* qmean,
		    struct OL_SET* pSet1, struct OL_SET* pSet2)
{
  unsigned int i;
  
  struct OL_SET Setd = ol_set_alloc (Set.size);

  double** v = ol_g_alloc ();
  double* theta = ol_e_alloc ();

  (*pSet1) = ol_set_alloc (Set.size);
  (*pSet2) = ol_set_alloc (Set.size);

  for (i = 0; i < Set.size; i++)
    ol_q_q_qdisori (qmean, Set.q[i], "cubic", Setd.q[i]);

  ol_set_aniso (Setd, v, theta);
      
  (*pSet1).size = 0;
  (*pSet2).size = 0;
  for (i = 0; i < Setd.size; i++)
  {
    ol_q_aniso_theta (Setd.q[i], v, theta);

    if (theta[0] < 0)
      ol_q_memcpy (Set.q[i], (*pSet1).q[(*pSet1).size++]);
    else
      ol_q_memcpy (Set.q[i], (*pSet2).q[(*pSet2).size++]);
  }

  ol_set_free (Setd);
  ol_g_free (v);
  ol_r_free (theta);

  return;
}

void
ol_set_aniso_v_modes (struct OL_SET Set, double* qmean, double** v,
		      struct OL_SET* pSet1, struct OL_SET* pSet2)
{
  ol_set_aniso_v_thetatr_modes (Set, qmean, v, 0, pSet1, pSet2);

  return;
}

void
ol_set_aniso_v_thetatr_modes (struct OL_SET Set, double* qmean, double** v,
	        double thetatr, struct OL_SET* pSet1, struct OL_SET* pSet2)
{
  unsigned int i;
  int* id = ut_alloc_1d_int (Set.size);
  int size1, size2;

  ol_set_aniso_v_thetatr_modes_id (Set, qmean, v, thetatr, id);
  
  size1 = ut_array_1d_int_nbofthisval (id, Set.size, 1);
  size2 = ut_array_1d_int_nbofthisval (id, Set.size, 2);
  (*pSet1) = ol_set_alloc (size1);
  (*pSet2) = ol_set_alloc (size2);
  
  (*pSet1).size = 0;
  (*pSet2).size = 0;
  for (i = 0; i < Set.size; i++)
  {
    if (id[i] == 1)
    {
      (*pSet1).size++;
      ol_q_memcpy (Set.q[i], (*pSet1).q[(*pSet1).size - 1]);
      (*pSet1).weight[(*pSet1).size - 1] = Set.weight[i];
    }
    else
    {
      (*pSet2).size++;
      ol_q_memcpy (Set.q[i], (*pSet2).q[(*pSet2).size - 1]);
      (*pSet2).weight[(*pSet2).size - 1] = Set.weight[i];
    }
  }

  ut_free_1d_int (id);

  return;
}

void
ol_set_aniso_v_thetatr_modes_id (struct OL_SET Set, double* qmean,
				 double** v, double thetatr, int* id)
{
  unsigned int i;
  struct OL_SET Setd = ol_set_alloc (Set.size);
  double* theta = ol_e_alloc ();

  for (i = 0; i < Set.size; i++)
    ol_q_q_qdisori (qmean, Set.q[i], "cubic", Setd.q[i]);

  for (i = 0; i < Setd.size; i++)
  {
    ol_q_aniso_theta (Setd.q[i], v, theta);

    id[i] = (theta[0] < thetatr) ? 1 : 2;
  }

  ol_set_free (Setd);
  ol_r_free (theta);

  return;
}
